/*
 * CompanyOwnershipProblem.java
 *
 * Solution to the "company ownership problem".
 *
 * Bartlomiej Brzozowiec (bartml @ wp.pl), IV/V 2011
 * License: GPLv3 or later (see http://www.gnu.org/licenses/gpl.html)
 */

/*
 * Analysis: the problem can be solved with graph algorithms. We can represent
 * the dataset as a directed weighted graph where vertices are companies, weights are
 * percentages of shares, and direction is an ownership relation.
 *
 * The solution is this: first we identify all companies that may be controlling other
 * companies (this will be our sourceVertices list). Then we use BFS (breadth-first
 * search) algorithm (see eg. "Introduction to Algorithms" by Thomas H. Cormen,
 * Charles E. Leiserson, Ronald L. Rivest, and Clifford Stein) for each vertex
 * from sourceVertices list to calculate sums of percentages for each vertex
 * reachable from the source. During this we also create another adjacency list
 * (to save space and number of operations, since the whole graph may be
 * not reachable from a given source). Then we modify this adjacency list by
 * replacing weight with sum of weights for each visited vertex. Then we run
 * BFS again, but this time we recurse only into arcs that have sum of weights
 * greater than CONTROLLING_THRESHOLD; we also record all visited vertices, which
 * gives us a final solution for given source vertex.
 * Why is it so complicated? Well, there is probably no simpler way.
 * Consider the following example:
 * Input:
 * 1	2	51
 * 2	5	20
 * 3	5	20
 * 4	5	9
 * 5	8	75
 * 6	7	70
 * 7	5	9
 * 1	3	52
 * 1	4	53
 * 4	6	60
 *
 * Proper output:
 * 1	2
 * 1	3
 * 1	4
 * 1	5
 * 1	6
 * 1	7
 * 1	8
 * 4	6
 * 4	7
 * 5	8
 * 6	7
 *
 */

package companyownershipproblem;

import java.io.*;
import java.util.*;

/**
 * @author Bartlomiej Brzozowiec (bartml @ wp.pl), IV/V 2011
 */
public class CompanyOwnershipProblem {
    // -----------------------------------------------------------------------
    // CONSTANTS
    // -----------------------------------------------------------------------
    /**
     * If sum of percentages of shares of given company which are in possession of another
     * company is greater than this value, then the company CONTROLS another company.
     */
    static final int CONTROLLING_THRESHOLD = 50;

    // -----------------------------------------------------------------------
    // VARIABLES
    // -----------------------------------------------------------------------
    /**
     * Adjacency list with data read directly from the user.
     */
    AdjacencyList<Integer, Integer> mainAdjList;

    /**
     * List of "source" vertices.
     */
    Set<Integer> sourceVertices;

    // -----------------------------------------------------------------------
    // non-static METHODS (including constructor)
    // -----------------------------------------------------------------------
    CompanyOwnershipProblem() {
        mainAdjList    = new AdjacencyList<Integer, Integer>();
        sourceVertices = new TreeSet<Integer>(); // it is automatically sorted
    }

    /**
     * Solve and print solution for Company Ownership Problem.
     */
    public void solveAndPrint() {
        // debug:
        //System.out.println( "solveAndPrint()" );
        //System.out.println( mainAdjList );
        //System.out.println( sourceVertices );

        for( Integer source : sourceVertices ) {
            // TreeSet is automatically sorted
            TreeSet<Integer> solution = solve( source );
            for( Integer s : solution ) {
                if( s == source ) { // we don't need to list source vertex - skip such case
                    continue;
                }
                System.out.println( source + "\t" + s );
            }
        }
    }

    /**
     * Solve "company ownership problem" for a given source vertex (ie. for
     * a single company).
     * @param source source vertex
     * @return list of companies controlled by a source company
     */
    private TreeSet solve( Integer source ) {
        // debug:
        //System.out.println( "solve()" );

        AdjacencyList<Integer, Integer> adjListForThisSource =
                new AdjacencyList<Integer, Integer>();
        Map<Integer, Integer> visitedVerticesWithSumOfWeights =
                new HashMap<Integer, Integer>();
        BFS1( source, adjListForThisSource, visitedVerticesWithSumOfWeights );

        // debug:
        //System.out.println( adjListForThisSource );
        //System.out.println( visitedVerticesWithSumOfWeights );

        // now, update adjacency list for this source -- put sums of weights in
        // proper places
        updateAdjList( source, adjListForThisSource, visitedVerticesWithSumOfWeights );

        // debug:
        //System.out.println( adjListForThisSource );
        //System.out.println( visitedVerticesWithSumOfWeights );

        // run BFS again, but this time only gather visited vertices;
        // follow an arc only if its weight is bigger than CONTROLLING_THRESHOLD
        return BFS2( source, adjListForThisSource );
    }

    /**
     * First BFS (breadth-first search). It gathers data needed for further use.<br />
     * <b>NOTE</b>: This method "returns" two data structures through modifying
     * its arguments!
     * @param source source vertex
     * @param adjList adjacency list for vertices reachable from given source<br />
     *        <b>NOTE</b>: pass a new empty non-anonymous object here, it
     *                     will be filled by this method
     * @param verticesWithSums list of visited vertices with sum of weights<br />
     *        <b>NOTE</b>: pass a new empty non-anonymous object here, it will be filled by this
     *                     method
     */
    private void BFS1( Integer source,
                       AdjacencyList<Integer, Integer> adjList,
                       Map<Integer, Integer> verticesWithSums ) {
        HashSet<Integer> visited = new HashSet<Integer>();
        LinkedList<Integer> queue = new LinkedList<Integer>();

        // mark source vertex as visited
        visited.add( source );

        queue.addLast( source );
        // main BFS loop
        while( ! queue.isEmpty() ) {
            Integer u = queue.removeFirst();
            for( Integer v : mainAdjList.getNeighbours( u ) ) {
                Integer weight = mainAdjList.getWeight( u, v );
                // build adjacency list for a given source only
                adjList.add_ijp( u, v, weight );
                // calculate sums of weights for visited vertices
                Integer sum = verticesWithSums.containsKey( v ) ?
                                verticesWithSums.get( v ) + weight : weight;
                verticesWithSums.put( v, sum );
                // if not visited...
                if( ! visited.contains( v ) ) {
                    visited.add( v ); // mark this vertex as visited
                    queue.addLast( v );
                }
            }
            visited.add( u );
        }
    }

    /**
     * Update given adjacency list with data from verticesWithSums for given
     * source vertex.<br />
     * <b>NOTE</b>: this method modifies adjList!
     * @param source
     * @param adjList
     * @param verticesWithSums
     */
    private void updateAdjList( Integer source,
                       AdjacencyList<Integer, Integer> adjList,
                       Map<Integer, Integer> verticesWithSums ) {
        for( Integer vertex : verticesWithSums.keySet() ) {
            Integer sum = verticesWithSums.get( vertex );
            // skip sums less than or equal to CONTROLLING_THRESHOLD -- there is
            // no point in updating them;
            // also we don't need to list source vertex -- see text of the task
            // ("...with h different from s.")
            if( sum <= CONTROLLING_THRESHOLD || vertex == source ) {
                continue;
            }
            // set sums of percentage of ownership in all relevant places
            // in adjacency list
            adjList.setAllNewWeight( vertex, sum );
        }
    }

    /**
     * Second BFS (breadth-first search). It traverses the graph represented in
     * adjacency list <code>adjList</code>, starting from source vertex
     * <code>source</code> and gathers visited vertices.<br />
     * This time it follows an arc only if its weight is bigger than CONTROLLING_THRESHOLD.
     * @param source source vertex
     * @param adjList adjacency list for vertices reachable from given source
     * @return list of visited vertices -- solution for a given source
     */
    private TreeSet BFS2( Integer source, AdjacencyList<Integer, Integer> adjList ) {
        HashSet<Integer> visited = new HashSet<Integer>();
        LinkedList<Integer> queue = new LinkedList<Integer>();

        // mark source vertex as visited
        visited.add( source );

        queue.addLast( source );
        // main BFS loop
        while( ! queue.isEmpty() ) {
            Integer u = queue.removeFirst();
            for( Integer v : adjList.getNeighbours( u ) ) {
                Integer sum = adjList.getWeight( u, v );
                // if not visited...
                // recurse into it only if sum of percentages is bigger than
                // CONTROLLING_THRESHOLD
                if( ! visited.contains( v ) && sum > CONTROLLING_THRESHOLD ) {
                    visited.add( v ); // mark this vertex as visited
                    queue.addLast( v );
                }
            }
            visited.add( u );
        }

        return new TreeSet<Integer>( visited ); // we want results sorted...
    }

    // -----------------------------------------------------------------------
    // static METHODS (including an entry point, ie. main())
    // -----------------------------------------------------------------------
    /**
     * Read data from file(s) given as command-line argument(s) or
     * from stdin if there is no command-line argument.
     * If the only argument is "-h" or "--help" it prints help and exits.
     * @param args the command line arguments
     * @throws IOException
     */
    public static void main(String[] args) throws IOException {
        // be helpful if user says so
        if( args.length == 1 && ( args[0].equalsIgnoreCase( "-h" ) ||
                args[0].equalsIgnoreCase( "--help" )) ) {
            System.err.println( "This program solves Company Ownership Problem.\n" +
              "It reads positive integers triples (i,j,p) which means that the company i owns p% of company j." );
            System.err.println( "The task is to write all the pairs (h,s) so that company h controls company s." );
            System.err.println( "It is said that a company A controls company B if," +
              " at least, one of the following conditions is satisfied:\n" +
              "* A owns more than " + CONTROLLING_THRESHOLD + "% of B\n" +
              "* The total sum of shares of B which are in possession of A and " +
              "companies controlled by A is greater than " + CONTROLLING_THRESHOLD + "%" );
            System.err.println( "\nData may come either from stdin or from file(s). " +
              "If there is no command-line argument, then program reads it from stdin, otherwise " +
              "it processes given files. Different datasets may be separated by an empty line." +
              "\n\nAuthor: Bartlomiej Brzozowiec <bartml @ wp.pl>, IV/V 2011" );

            System.exit( 1 );
        }

        BufferedReader in = null;
        try {
            if( args.length > 0 ) {
                for( String file : args ) {
                    in = new BufferedReader( new FileReader( file ) );
                    System.err.println( "==> Output for file \"" + file + "\" <==" );
                    processInput( in );
                    in = null;
                }
            } else {
                in = new BufferedReader( new InputStreamReader( System.in ) );
                processInput( in );
            }
        } finally {
            if( in != null ) {
                in.close();
            }
        }
    }

    /**
     * Build adjacency list from given data, also gather "source" vertices.
     * Then solve the problem for given dataset.
     * @param in object related with stdin or file, depending on command-line
     *           arguments
     */
    private static void processInput( BufferedReader in ) throws IOException {
        CompanyOwnershipProblem cop = new CompanyOwnershipProblem();
        String s = null;
        while( (s = in.readLine()) != null ) {
            if( s.trim().equals( "" ) ) {   // empty line -- new dataset
                cop.solveAndPrint();
                // separate different solutions with a blank line
                System.out.println();
                cop = new CompanyOwnershipProblem();
                continue;
            }
            Scanner scanner = new Scanner( s );
            int i = scanner.nextInt();
            int j = scanner.nextInt();
            int p = scanner.nextInt();
            cop.mainAdjList.add_ijp( i, j, p );

            // gather the list of companies that control another company(s)
            // (the bigger problem is to find exactly which companies are
            // controlled...)
            if( p > CONTROLLING_THRESHOLD ) {
                cop.sourceVertices.add( i );
            }
        }
        cop.solveAndPrint();
    }
}

// ----------------------------------------------------------------------------
/**
 * This class effectively implements hash of hashes, needed to represent
 * a directed graph with weights (which is needed to solve "company ownership
 * problem") in an adjacency list representation.
 * @author bb
 * @param <EIJ> type of i and j -- identifiers of companies (keys of a hashes)
 * @param <EP>  type of p -- percentage of ownership (values in an inner hash)
 */
class AdjacencyList <EIJ, EP> {
    private Map<EIJ, Map<EIJ, EP>> adjList;

    public AdjacencyList() {
        adjList = new HashMap<EIJ, Map<EIJ, EP>>();
    }

    /**
     * Adds vertex <i>j</i> with weight <i>p</i> to adjacency list of vertex <i>i</i>.
     * We need this method to easily add an entry from a given dataset.
     * @param i company identifier
     * @param j company identifier
     * @param p percentage of ownership
     */
    public void add_ijp( EIJ i, EIJ j, EP p ) {
        Map<EIJ, EP> el = adjList.get( i );
        if( el == null ) {  // new key
            adjList.put( i, el = new HashMap<EIJ, EP>() );
        }
        el.put( j, p );
    }

    /**
     * @param vertex
     * @return set of neighbours of given vertex or empty set if there aren't
     *         any neighbours
     */
    public Set<EIJ> getNeighbours( EIJ vertex ) {
        return( adjList.containsKey( vertex ) ?
            adjList.get( vertex ).keySet() :
            new HashSet<EIJ>() );
    }
    
    /**
     * @param u
     * @param v
     * @return weight (value of arc in weighted graph) for arc connecting u and v
     */
    public EP getWeight( EIJ u, EIJ v ) {
        return adjList.get( u ).get( v );
    }

    /**
     * Update adjacency list for given vertex, ie. set weight to given weight
     * in all relevant places.
     * @param vertex
     * @param newWeight
     */
    public void setAllNewWeight( EIJ vertex, EP newWeight ) {
        for( EIJ key : adjList.keySet() ) {
            Map<EIJ, EP> el = adjList.get( key );
            for( Map.Entry<EIJ, EP> e : el.entrySet() ) {
                if( e.getKey() == vertex ) {
                    e.setValue( newWeight );
                }
            }
        }
    }

    // ------
    /**
     * Overloaded method (every Object has it), useful mainly for debugging purposes.
     * @return adjacency list of this object
     */
    @Override
    public String toString() {
        String s = new String();
        for( EIJ key : adjList.keySet() ) {
            s += key + ": ";
            Map<EIJ, EP> el = adjList.get( key );
            for( Map.Entry<EIJ, EP> e : el.entrySet() ) {
                s += "(" + e.getKey() + "->" + e.getValue() + ") ";
            }
            s += "\n";
        }
        return s;
    }
}
// ----------------------------------------------------------------------------
